"""
Helper functions for Cypher AI
"""
from typing import Union, Type, Any, get_type_hints

from pydantic import BaseModel

import functools
import inspect
import re


def pydantic_model_to_str(model: Union[BaseModel, Type[BaseModel]]) -> str:
    """
    Converts a pydantic model to a string representation of the class definition

    :param model:
    :return:
    """

    def get_field_def(name, field_info) -> str:
        """
        Returns a string representation of a field definition

        :param name:
        :param field_info:
        :return:
        """
        default = field_info.default if not field_info.required else "..."
        description = field_info.field_info.description
        if description:
            description = ', description="' + description.replace("\n", " ") + '"'
        else:
            description = ""
        return f'{name}: {field_info.outer_type_.__name__} = Field({default}{description})'

    fields = [get_field_def(name, field_info) for name, field_info in model.__fields__.items()]

    docstring = f'"""\n{model.__doc__}\n"""' if model.__doc__ else ""
    class_def = f'class {model.__name__}(BaseModel):\n{docstring}'
    class_fields = '\n'.join([f'    {line}' for line in fields])
    return f'{class_def}\n{class_fields}'


def pydantic_models_to_str(models: list[Union[BaseModel, Type[BaseModel]]]) -> str:
    """
    Converts a list of pydantic models to a string representation of the class definitions
    :param models:
    :return:
    """
    return '\n\n'.join([pydantic_model_to_str(model) for model in models])


def class_definition_in_str_for_dataclasses(model: Any):
    """
    Creates schema for dataclass

    :param model:
    :return:
    """
    fields = ', '.join([f'{name}: {field_type.__name__}'
                        for name, field_type in get_type_hints(model).items()])

    docstring = f'"""\n{model.__doc__}\n"""' if model.__doc__ else ""
    class_def = f'@dataclasses.dataclass\nclass {model.__name__}:\n{docstring}'
    class_fields = '\n'.join([f'    {line}' for line in fields.split(', ')])
    return f'{class_def}\n{class_fields}'


# Below functions are taken from: https://github.com/amikos-tech/funkagent/blob/feature/partials/funkagent/parser.py


def type_mapping(dtype: type) -> str:
    """
    Maps python types to OpenAPI types

    :param dtype:
    :return:
    """
    if dtype == float:
        return "number"
    elif dtype == int:
        return "integer"
    elif dtype == str:
        return "string"
    else:
        return "string"


def extract_params(doc_str: str) -> dict[str, str]:
    """
    Extracts parameters from docstring
    :param doc_str:
    :return:
    """
    # split doc string by newline, skipping empty lines
    params_str = [line for line in doc_str.split("\n") if line.strip()]
    params = {}
    for line in params_str:
        # we only look at lines starting with ':param'
        if line.strip().startswith(':param'):
            param_match = re.findall(r'(?<=:param )\w+', line)
            if param_match:
                param_name = param_match[0]
                desc_match = line.replace(f":param {param_name}:", "").strip()
                # if there is a description, store it
                if desc_match:
                    params[param_name] = desc_match
    return params


def func_to_json(func):
    """
    Converts a function to a json schema for OpenAI

    :param func:
    :return:
    """
    # Check if the function is a functools.partial
    if isinstance(func, functools.partial) or isinstance(func, functools.partialmethod):
        fixed_args = func.keywords
        _func = func.func
    else:
        fixed_args = {}
        _func = func

    # first we get function name
    func_name = _func.__name__
    # then we get the function annotations
    argspec = inspect.getfullargspec(_func)
    # get the function docstring
    func_doc = inspect.getdoc(_func)
    # parse the docstring to get the description
    func_description = ''.join([line for line in func_doc.split("\n") if not line.strip().startswith(':')])
    # parse the docstring to get the descriptions for each parameter in dict format
    param_details = extract_params(func_doc) if func_doc else {}
    # attach parameter types to params and exclude fixed args
    # get params
    params = {}
    for param_name in argspec.args:
        if param_name not in fixed_args.keys():
            params[param_name] = {
                "description": param_details.get(param_name) or "",
                "type": type_mapping(argspec.annotations.get(param_name, type(None)))
            }
    # calculate required parameters excluding fixed args
    # _required = [arg for arg in argspec.args if arg not in fixed_args]
    _required = [i for i in argspec.args if i not in fixed_args.keys()]
    if inspect.getfullargspec(_func).defaults:
        _required = [argspec.args[i] for i, a in enumerate(argspec.args) if
                     argspec.args[i] not in inspect.getfullargspec(_func).defaults and argspec.args[
                         i] not in fixed_args.keys()]
    # then return everything in dict
    return {
        "name": func_name,
        "description": func_description,
        "parameters": {
            "type": "object",
            "properties": params
        },
        "required": _required
    }



"""
Main module for Cypher AI Helper
"""
import json
import logging
import os
from time import sleep
from typing import Any

import openai
from neo4j import Driver
from neo4j.exceptions import CypherSyntaxError

from cypher_ai_helper.utils import func_to_json

curdir = os.path.dirname(os.path.realpath(__file__))

logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=logging.DEBUG)
logger = logging.getLogger(__name__)


def _send_chat_messages(messages, _model: str = "gpt-3.5-turbo-0613", functions=None) -> Any:
    """
    Send chat messages to the OpenAI API

    :param messages: Messages to send
    :param _model: Model to use
    :param functions: Functions to use
    :return: Response from the API
    """
    max_retry = 7
    retry = 0
    while True:
        try:
            if functions:
                response = openai.ChatCompletion.create(
                    model=_model,
                    messages=messages,
                    functions=functions if functions else [],
                    # auto is default, but we'll be explicit,
                    function_call="auto",
                    # we don't want creativity but consistency
                    temperature=0.0,
                    top_p=1.0,
                    frequency_penalty=0.0,
                    presence_penalty=0.0,
                    max_tokens=256,
                )
            else:
                response = openai.ChatCompletion.create(
                    model=_model,
                    messages=messages,
                    temperature=0.0,
                    top_p=1.0,
                    frequency_penalty=0.0,
                    presence_penalty=0.0,
                    max_tokens=256,
                )
            # decode utf-8 bytes to unicode
            return response["choices"][0]["message"]
        except Exception as oops:
            logger.warning(f'\n\nError communicating with OpenAI: "{oops}"')
            if 'maximum context length' in str(oops):
                a = messages.pop(1)
                logger.warning(f'\n\n DEBUG: Trimming oldest message: {a}')
                continue
            retry += 1
            if retry >= max_retry:
                logger.error(f"\n\nExiting due to excessive errors in API: {oops}")
                raise oops
            logger.warning(f'\n\nRetrying in {2 ** (retry - 1) * 5} seconds...')
            sleep(2 ** (retry - 1) * 5)


def execute_query(query: str, driver: Driver, database: str, params: dict[str, Any] = None) -> list[any]:
    """
    Executes Neo4j query and returns all results

    :param query: Query to run
    :param driver: Driver to use in Neo4j
    :param database: Database to use in Neo4j
    :param params: Parameters to use in the query
    :return:
    """
    _results = []
    with driver.session(database=database) as session:
        logger.debug(f"Executing query: {query} with params: {params}")
        result = session.run(query=query, parameters=params)
        for record in result:
            _results.append(record)
    return _results


query_cache = {}


def cypher_query(query_id: str, schema: str, instruction: str, input_vars: list[str], model="gpt-3.5-turbo-0613",
                 overwrite_cache: bool = False, func: callable = None) -> tuple[str, Any]:
    """
    Generate a Cypher query based on schema, input data and a question

    :param query_id: Query ID - the query ID to store in cache
    :param schema: Schema of the data - the schema of the data to use in the Cypher query
    :param instruction: Instruction from the user
    :param input_vars: Input variables - list of input variables to use in the Cypher query
    :param model: Model to use - OpenAI Model to use
    :param overwrite_cache: Overwrite cache - boolean flag to overwrite cache (e.g. if the query by id exits in the \
        cache it is removed and a new query is generated)
    :param func: Function to use to test the query
    :return:
    """
    global query_cache

    if overwrite_cache:
        del query_cache[query_id]

    if query_id in query_cache:
        logger.debug(f"Returning cached query for {query_id}")
        return query_cache[query_id]

    with open(os.path.join(curdir, "system_prompt"), "r") as f:
        _system_prompt = f.read()
    with open(os.path.join(curdir, "user_prompt"), "r") as f:
        _prompt = f.read()
    _prompt = _prompt.replace("{{schema}}", schema)
    _prompt = _prompt.replace("{{instruction}}", instruction)
    _prompt = _prompt.replace("{{input_vars}}", ", ".join(input_vars))

    _messages = [
        {"role": "system", "content": _system_prompt},
        {"role": "user", "content": _prompt},
    ]
    _functions = None

    _response_message = _send_chat_messages(_messages, _model=model, functions=_functions)
    _messages.append(json.loads(json.dumps(_response_message, ensure_ascii=False)))
    logger.debug(f"Messages: {_messages}")
    _follow_up_messages = _messages
    if func:
        _new_messages = []
        _functions = [func_to_json(func)]
        assert len(_functions[0]["parameters"]) > 0, "Function must have at least one parameter"
        _new_messages.append(
            {
                "role": "user",
                "content": f"Execute the the following query: {_response_message['content']}\n\n"
                           f"Do not attempt to generate values for placeholders, just execute the query as is.",
            }
        )
        _response_message = _send_chat_messages(_new_messages, _model=model, functions=_functions)
        _new_messages.append(json.loads(json.dumps(_response_message, ensure_ascii=False)))
        logger.debug(f"Messages: {_new_messages}")
        _follow_up_messages = _new_messages

    while "function_call" in _response_message:
        _function_to_call = func
        logger.info(f"Function to call: {_functions[0]['name']}")
        _function_args = json.loads(_response_message["function_call"]["arguments"])
        try:
            _func_response = _function_to_call(**_function_args)
            _follow_up_messages.append(
                {
                    "role": "function",
                    "name": _functions[0]["name"],
                    "content": f"{_func_response}",
                }
            )
            # Note: the below is a bit dumb but works for now
            return next(iter(_function_args.items())), _func_response
        except CypherSyntaxError as e:
            logger.error(f"Error executing query function: {e}")
            _follow_up_messages.append(
                {
                    "role": "function",
                    "name": _functions[0]["name"],
                    "content": f"Error executing query function: {repr(e)}",
                })
            _follow_up_messages.append({
                "role": "user",
                "content": "The query failed. Please consider the error and rewrite the query."
            })
            _response_message = _send_chat_messages(_follow_up_messages, _model=model, functions=_functions)

    logger.debug(f"Messages: {_follow_up_messages}")
    logger.debug(f"LLM Query Response: {_response_message}")
    query_cache[query_id] = _response_message["content"]
    return _response_message["content"], None
